/*
    Copyright 2010 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#define MAINPROG
#include "disasm.h"

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1.1"



int check_overwrite(u8 *fname);
u8 *fdload(u8 *fname, int *fsize);
int dump2func(u8 *buff, int buffsz, FILE *fdo);
void std_err(void);



int main(int argc, char *argv[]) {
    FILE    *fdo = stdout;
    int     buffsz;
    u8      *fname,
            *fout,
            *buff;

    setbuf(stdout, NULL);

    fputs("\n"
        "dump2func ripped functions skeleton generator "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "  disassembler engine from Oleh Yuschuk\n"
        "\n", stderr);

    if(argc < 2) {
        fprintf(stderr, "\n"
            "Usage: %s <input_dump> [output.c]\n"
            "\n", argv[0]);
        exit(1);
    }
    fname = argv[1];

    buff = fdload(fname, &buffsz);

    if(argc > 2) {
        fout = argv[2];
        if(check_overwrite(fout) < 0) exit(1);
        fdo = fopen(fout, "wb");
        if(!fdo) std_err();
    }

    dump2func(buff, buffsz, fdo);

    if(fdo != stdout) fclose(fdo);
    fprintf(stderr, "- done\n");
    return(0);
}



u8 *fdload(u8 *fname, int *fsize) {
    struct stat xstat;
    FILE    *fd;
    int     size;
    u8      *buff;

    fd = fopen(fname, "rb");
    if(!fd) std_err();
    fstat(fileno(fd), &xstat);
    size = xstat.st_size;
    buff = malloc(size);
    if(!buff) std_err();
    fread(buff, 1, size, fd);
    fclose(fd);
    if(fsize) *fsize = size;
    return(buff);
}



int check_overwrite(u8 *fname) {
    FILE    *fd;
    u8      ans[16];

    if(!fname) return(0);
    fd = fopen(fname, "rb");
    if(!fd) return(0);
    fclose(fd);
    fprintf(stderr, "- the file \"%s\" already exists\n  do you want to overwrite it (y/N/all)? ", fname);
    fgets(ans, sizeof(ans), stdin);
    if(tolower(ans[0]) == 'y') return(0);
    return(-1);
}



int dump2func(u8 *buff, int buffsz, FILE *fdo) {
typedef struct {
    u32     addr;
    u32     offset;
    u32     variable;
    u8      *instr;
} list_t;

    list_t  *list = NULL;
    t_disasm da;
    int     i,
            j,
            asm_size,
            addr,
            offset,
            list_num,
            variables      = 0,
            *variable      = NULL,
            *variable_size = NULL;
    u8      *callconv;

    lowercase   = 1;
    extraspace  = 1;
    showmemsize = 1;

    list_num = buffsz / 5;  // quick and fast
    list = calloc(list_num, sizeof(list_t));
    if(!list) std_err();

    list_num = 0;
    for(addr = 0; addr < buffsz; addr += asm_size) {
        asm_size = Disasm(buff + addr, buffsz - addr, 0, &da, DISASM_CODE); // DISASM_DATA);
        if(asm_size <= 0) break;

        if(!(
            ((da.cmdtype & C_TYPEMASK) == C_CMD) ||
            ((da.cmdtype & C_TYPEMASK) == C_PSH) ||
            ((da.cmdtype & C_TYPEMASK) == C_DAT) ||
            ((da.cmdtype & C_TYPEMASK) == C_JMP))) {
            continue;
        }

        for(i = 0; i < 2; i++) {
            switch(i) {
                case 0: offset = da.adrconst;   break;  // mov eax, dword ptr [4*ecx+OFFSET]
                case 1: offset = da.immconst;   break;  // like mov eax, OFFSET
                default: break;
            }
            if(offset <= 0) continue;
            if(offset <= 0x00400000) continue;
            break;
        }
        if(i >= 2) continue;

        for(i = 0; i <= (asm_size - 4); i++) {
            if(!memcmp(buff + addr + i, &offset, 4)) break;
        }

        list[list_num].addr     = addr + i;
        list[list_num].offset   = offset;
        list[list_num].variable = offset;
        list[list_num].instr    = strdup(da.result);
        list_num++;
    }

    for(i = 0; i < list_num; i++) {
        for(j = 0; j < list_num; j++) {
            if(list[j].variable > list[i].variable) continue;
            if((list[j].variable + 0x00040000) < list[i].variable) continue;
            list[i].variable = list[j].variable;
        }
    }

    fprintf(fdo, "#include <stdio.h>\n");
    fprintf(fdo, "#include <stdlib.h>\n");
    fprintf(fdo, "#include <string.h>\n");
    fprintf(fdo, "#ifdef WIN32\n");
    fprintf(fdo, "    #include <windows.h>\n");
    fprintf(fdo, "#endif\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");

    if(list_num) {
        fprintf(fdo, "// variables and buffers guessed by the tool\n");
        variable = calloc(list_num, sizeof(int));
        if(!variable) std_err();
        variable_size = calloc(list_num, sizeof(int));
        if(!variable_size) std_err();
        for(i = 0; i < list_num; i++) {
            for(j = 0; variable[j]; j++) {
                if(variable[j] == list[i].variable) {
                    variable_size[j] = list[i].offset - list[i].variable;
                    break;
                }
            }
            if(variable[j]) continue;
            fprintf(fdo, "static unsigned char *CHANGEME_%08x = NULL;\n", list[i].variable);
            variable[j] = list[i].variable;
            variable_size[j] = 0;
        }
        fprintf(fdo, "\n");
        fprintf(fdo, "\n");
        fprintf(fdo, "\n");
        for(i = 0; variable[i]; i++);
        variables = i;
    }

    fprintf(fdo, "static /*const*/ unsigned char CHANGEME_dump[] = {");
    for(i = 0; i < buffsz; i++) {
        if(!(i & 15)) fprintf(fdo, "\n    ");
        fprintf(fdo, "0x%02x,", buff[i]);
    }
    fprintf(fdo, "/*INT3*/ 0xcc };\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");

    callconv = "__stdcall"; // try to guess, not necessary
    for(i = buffsz - 1; i >= 0; i--) {
        if(buff[i] == 0x90) continue;
        if(buff[i] == 0xcc) continue;
        if(buff[i] == 0x00) continue;   // mah
        if(buff[i] == 0xc3) callconv = "__cdecl";
        break;
    }

    fprintf(fdo, "int (%s *CHANGEME)(void) = NULL; \\\\EDIT create the exact prototype including all the needed arguments\n", callconv);
    fprintf(fdo, "    // if you need to use dumped functions that use an unsupported calling convention (like safecall\n");
    fprintf(fdo, "    // or borland fastcall) take a look at the file calling_conventions.h in my QuickBMS project\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "// anti DEP limitation, allocation is necessary\n");
    fprintf(fdo, "void *CHANGEME_alloc(void *dump, int dumpsz) {\n");
    fprintf(fdo, "    int     pagesz;\n");
    fprintf(fdo, "    void    *ret;\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "    pagesz = (dumpsz + 4095) & (~4095); // useful for pages? mah\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "#ifdef WIN32\n");
    fprintf(fdo, "    ret = VirtualAlloc(\n");
    fprintf(fdo, "        NULL,\n");
    fprintf(fdo, "        pagesz,\n");
    fprintf(fdo, "        MEM_COMMIT | MEM_RESERVE,\n");
    fprintf(fdo, "        PAGE_EXECUTE_READWRITE);    // write for memcpy\n");
    fprintf(fdo, "#else\n");
    fprintf(fdo, "    ret = malloc(pagesz);\n");
    fprintf(fdo, "    mprotect(\n");
    fprintf(fdo, "        ret,\n");
    fprintf(fdo, "        pagesz,\n");
    fprintf(fdo, "        PROT_EXEC | PROT_WRITE);    // write for memcpy\n");
    fprintf(fdo, "#endif\n");
    fprintf(fdo, "    memcpy(ret, dump, dumpsz);\n");
    fprintf(fdo, "    return(ret);\n");
    fprintf(fdo, "}\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");
    fprintf(fdo, "\n");

    fprintf(fdo, "void CHANGEME_init(void) {\n");
    fprintf(fdo, "    if(!CHANGEME) {\n");
    fprintf(fdo, "\n");

    if(variables) {
        fprintf(fdo, "        \\\\EDIT check the size of the following variables (so remove 0x2000) and remove the false positives\n");
        for(i = 0; i < variables; i++) {
            fprintf(fdo, "        CHANGEME_%08x = calloc(0x%x + 0x2000, 1);\n", variable[i], variable_size[i] + 4);
        }
        fprintf(fdo, "\n");
    }

    //fprintf(fdo, "        CHANGEME = (void *)(CHANGEME_dump + 0x00000000);   \\\\EDIT init function, check the offset\n");
    fprintf(fdo, "        CHANGEME = CHANGEME_alloc(CHANGEME_dump, sizeof(CHANGEME_dump));\n");
    fprintf(fdo, "\n");

    if(list_num) {
        fprintf(fdo, "        #define PATCHIT(X,Y) *(int *)((char *)CHANGEME + X) = (int)(Y);\n");
        for(i = 0; i < list_num; i++) {
            fprintf(fdo, "        PATCHIT(0x%08x, CHANGEME_%08x + 0x%08x) // %s\n", list[i].addr, list[i].variable, list[i].offset - list[i].variable, list[i].instr);
        }
        fprintf(fdo, "        #undef PATCHIT\n");
        fprintf(fdo, "\n");
    }

    fprintf(fdo, "        \\\\EDIT add your other custom initializations here writing on CHANGEME, example\n");
    fprintf(fdo, "        \\\\EDIT     unsigned char *p = (unsigned char *)CHANGEME;\n");
    fprintf(fdo, "        \\\\EDIT     p[100] = 0x90; p[101] = 0xe8; *(int *)(p + 102) = 0x12345678;\n");
    fprintf(fdo, "\n");

    fprintf(fdo, "    }\n");
    fprintf(fdo, "    // everything else that needs to be initialized all the times must be placed here\n");
    fprintf(fdo, "    // like for example the zeroing of a buffer or the initialization of a variable\n");
    fprintf(fdo, "\n");

    fprintf(fdo, "}\n");
    fprintf(fdo, "\n");

    fprintf(fdo, "    \\\\EDIT remember to call EVER CHANGEME_init before the function ALL the times:\n");
    fprintf(fdo, "    \\\\EDIT   CHANGEME_init();\n");
    fprintf(fdo, "    \\\\EDIT   ret = CHANGEME(arg1, arg2, arg3);\n");
    fprintf(fdo, "\n");

    free(list);
    return(0);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


