/*
    Copyright 2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <windows.h>
    #define FILE_CRLF   "\r\n"  // change it to \n if you don't want \r\n used in strings.txt and stdout
#else
    #include "pe_nonwin.h"
    #define FILE_CRLF   "\n"    // as above, default *nix is line-feed only... and this is also the one I like
#endif

#define MAINPROG
#include "disasm.h"

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER             "0.2.3"
#define BUFFSZ          2000
#define MYSECTION       "stringz"
#define MYLINESDIV      '-'
#define STDINSZ         1000000
#define IS_WRONG        0       // this is the "type" field
#define IS_ASCII        1
#define IS_ASIAN        2
#define IS_UNICODE      4
#define SHF_WRITE       1
#define SHF_ALLOC       2
#define SHF_EXECINSTR   4
#define LINE_LF         1       // supported
#define LINE_CRLF       2       // supported
#define LINE_CR         3
#define LINE_LFCR       4



typedef struct {
    u8      e_ident[16];
    u16     e_type;
    u16     e_machine;
    u32     e_version;
    u32     e_entry;
    u32     e_phoff;
    u32     e_shoff;
    u32     e_flags;
    u16     e_ehsize;
    u16     e_phentsize;
    u16     e_phnum;
    u16     e_shentsize;
    u16     e_shnum;
    u16     e_shstrndx;
} elf_header_t;

typedef struct {
    u32     sh_name;
    u32     sh_type;
    u32     sh_flags;
    u32     sh_addr;     
    u32     sh_offset;
    u32     sh_size;
    u32     sh_link;
    u32     sh_info;
    u32     sh_addralign;
    u32     sh_entsize;
} elf_section_t;

typedef struct {
    u8      name[IMAGE_SIZEOF_SHORT_NAME + 1];  // yes I know the ELF have bigger names but this is used only to see if "stringz" already exists
    u32     rva;
    u32     rvasize;
    u32     offset;
    u32     size;
    u32     type;
} section_t;

typedef struct {
    int     offsets;            // number of same offsets for a specific string, so in how much places it's already pointed
    u32     *pointer_offset;    // important
    u32     *string_offset;     // mainly for visualization only
    u32     *string_rva;        // visualization only
    int     string_type;        // string_type and string are the only unique values
    u8      *string;            // so if an executable has 1000 identical strings at different offsets this tool will consider them one string only... cool
} strindex_t;

static const u8 MYLINES[] = "________________________________________________________________________________";
section_t   *section    = NULL;
strindex_t  *strindex   = NULL;
u32     filememsz       = 0,
        imagebase       = 0,
        entrypoint      = 0,
        sections        = 0,
        sections_align  = 0;
int     new_section_sz  = 0,
        new_section_rva = 0,
        asian_only      = 0,
        try_unicode     = 1,    // get unicode strings too by default: 68 00 65 00 6c 00 6c 00 6f 00 00 00
        minstringlen    = 4,    // 4 is the minimum for avoiding false positives
        force_crlf      = 0,    // not used
        c_string        = 0,
        no_filter       = 0,
        quiet           = 0,
        strindexsz      = 0,
        strindexn       = 0,
        binary_mode     = 0,
        align_string    = 0,
        scan_entryp     = 0;
u8      *filemem        = NULL,
        *new_section    = NULL;



void free_all(void) {
    if(new_section) free(new_section);
    if(section) free(section);
    if(filemem) free(filemem);
}



void std_err(u8 *err) {
    if(!err) err = strerror(errno);
    fprintf(stderr, "\nError: %s\n", err);
    free_all();
    exit(1);
}



u32 file2rva(u32 file) {
    u32     diff;
    int     i,
            ret;

    ret  = -1;
    diff = -1;
    for(i = 0; i < sections; i++) {
        if((file >= section[i].offset) && (file < (section[i].offset + section[i].size))) {
            if((file - section[i].offset) < diff) {
                diff = file - section[i].offset;
                ret  = i;
            }
        }
    }
    if(ret < 0) return(-1);
    return(imagebase + section[ret].rva + file - section[ret].offset);
}



u32 rva2file(u32 va) {
    u32     diff;
    int     i,
            ret;

    va  -= imagebase;
    ret  = -1;
    diff = -1;
    for(i = 0; i < sections; i++) {
        if(!section[i].rva) continue;
        if((va >= section[i].rva) && (va < (section[i].rva + section[i].rvasize))) {
            if((va - section[i].rva) < diff) {
                diff = va - section[i].rva;
                ret  = i;
            }
        }
    }
    if(ret < 0) return(-1);
    return(section[ret].offset + va - section[ret].rva);
}



int parse_PE(int add_new_section) {
    IMAGE_DOS_HEADER        *doshdr;
    IMAGE_NT_HEADERS        *nthdr;
    IMAGE_SECTION_HEADER    *sechdr;
    int     i,
            t;
    u8      *p;

    if(binary_mode) {
        imagebase      = 0;
        sections       = 1;
        sections_align = 0;

        section = realloc(section, sizeof(section_t) * sections);
        if(!section) std_err(NULL);

        strncpy(section[0].name, "", IMAGE_SIZEOF_SHORT_NAME);
        section[0].rva     = 0;
        section[0].rvasize = filememsz;
        section[0].offset  = 0;
        section[0].size    = filememsz;
        section[0].type    = 0;
        return(0);
    }

    if(add_new_section) {   // reallocate now or we lost the various pointers
        fprintf(stderr, "- add new section at the end of the executable from offset %08x\n", filememsz);
        filemem = realloc(filemem, filememsz + new_section_sz); // new_section_sz is ALREADY aligned!
        if(!filemem) std_err(NULL);
        memcpy(filemem + filememsz, new_section, new_section_sz);
    }

    p = filemem;
    doshdr  = (IMAGE_DOS_HEADER *)p;    p += sizeof(IMAGE_DOS_HEADER);
                                        p += doshdr->e_lfanew - sizeof(IMAGE_DOS_HEADER);
    nthdr   = (IMAGE_NT_HEADERS *)p;    p += sizeof(IMAGE_NT_HEADERS);

    if((doshdr->e_magic != IMAGE_DOS_SIGNATURE) || (nthdr->Signature != IMAGE_NT_SIGNATURE) || (nthdr->OptionalHeader.Magic != IMAGE_NT_OPTIONAL_HDR_MAGIC)) {
        return(-1);
    }
    imagebase      = nthdr->OptionalHeader.ImageBase;
    entrypoint     = imagebase + nthdr->OptionalHeader.AddressOfEntryPoint;
    sechdr         = (IMAGE_SECTION_HEADER *)p;
    sections       = nthdr->FileHeader.NumberOfSections;
    sections_align = nthdr->OptionalHeader.SectionAlignment;

    if(!new_section_rva) {
        for(i = 0; i < sections; i++) {
            t = sechdr[i].VirtualAddress + sechdr[i].Misc.VirtualSize;
            if(t > new_section_rva) new_section_rva = t;
        }
        t = new_section_rva % sections_align;
        if(t) new_section_rva += sections_align - t;
    }

    if(add_new_section) {
        nthdr->OptionalHeader.SizeOfImage += new_section_sz;
        nthdr->FileHeader.NumberOfSections = sections + 1;

        strncpy(sechdr[sections].Name, MYSECTION, IMAGE_SIZEOF_SHORT_NAME);
        sechdr[sections].Misc.VirtualSize       = new_section_sz;
        sechdr[sections].VirtualAddress         = new_section_rva;
        sechdr[sections].SizeOfRawData          = new_section_sz;
        sechdr[sections].PointerToRawData       = filememsz;
        sechdr[sections].PointerToRelocations   = 0;
        sechdr[sections].PointerToLinenumbers   = 0;
        sechdr[sections].NumberOfRelocations    = 0;
        sechdr[sections].NumberOfLinenumbers    = 0;
        sechdr[sections].Characteristics        = IMAGE_SCN_MEM_SHARED | IMAGE_SCN_MEM_READ | IMAGE_SCN_MEM_WRITE | IMAGE_SCN_CNT_INITIALIZED_DATA; // IMAGE_SCN_CNT_CODE | IMAGE_SCN_MEM_EXECUTE
        sections++;

        filememsz += new_section_sz;
    }

    section = realloc(section, sizeof(section_t) * sections);
    if(!section) std_err(NULL);

    for(i = 0; i < sections; i++) {
        strncpy(section[i].name, sechdr[i].Name, IMAGE_SIZEOF_SHORT_NAME);
        section[i].name[IMAGE_SIZEOF_SHORT_NAME] = 0;
        if(!add_new_section && !strcmp(section[i].name, MYSECTION)) {
            std_err("this executable already contains a \""MYSECTION"\" section made by this tool");
        }
        section[i].rva     = sechdr[i].VirtualAddress;
        section[i].rvasize = sechdr[i].Misc.VirtualSize;
        section[i].offset  = sechdr[i].PointerToRawData;
        section[i].size    = sechdr[i].SizeOfRawData;
        section[i].type    = sechdr[i].Characteristics;
        if(!section[i].rvasize) section[i].rvasize = section[i].size;    // Watcom
    }
    return(0);
}



int parse_ELF(int add_new_section) {
    elf_header_t    *elf_header;
    elf_section_t   *sechdr;
    int     i;
    u8      *p;

    if(add_new_section) {
        std_err("at the moment is not possible to put the strings back in an ELF executable\n");
    }

    p = filemem;
    elf_header = (elf_header_t *)p;     p += sizeof(elf_header_t);
    if(memcmp(elf_header->e_ident, "\x7f""ELF", 4)) return(-1);
    if(elf_header->e_ident[4] != 1) return(-1); // only 32 bit supported
    if(elf_header->e_ident[5] != 1) return(-1); // only little endian

    imagebase      = 0;
    entrypoint     = elf_header->e_entry;
    sechdr         = (elf_section_t *)(filemem + elf_header->e_shoff);
    sections       = elf_header->e_shnum;
    sections_align = 4096;  // is this good ???

    section = realloc(section, sizeof(section_t) * sections);   // in case in future will be added support for adding sections
    if(!section) std_err(NULL);

    for(i = 0; i < sections; i++) {
        strncpy(section[i].name, filemem + sechdr[elf_header->e_shstrndx].sh_offset + sechdr[i].sh_name, IMAGE_SIZEOF_SHORT_NAME);
        section[i].name[IMAGE_SIZEOF_SHORT_NAME] = 0;
        if(!add_new_section && !strcmp(section[i].name, MYSECTION)) {
            std_err("this executable already contains a \""MYSECTION"\" section made by this tool");
        }
        section[i].rva     = sechdr[i].sh_addr;
        section[i].rvasize = sechdr[i].sh_size;
        section[i].offset  = sechdr[i].sh_offset;
        section[i].size    = sechdr[i].sh_size;
        section[i].type    = sechdr[i].sh_flags;
        if(section[i].type & SHF_EXECINSTR) section[i].type = IMAGE_SCN_MEM_EXECUTE | IMAGE_SCN_CNT_CODE;   // windows compatible
        if(!section[i].rvasize) section[i].rvasize = section[i].size;    // Watcom
    }
    return(0);
}



FILE *load_file(u8 *fname, int openonly) {
    struct stat xstat;
    int     len;
    FILE    *fd;

    if(!strcmp(fname, "-")) {
        fd = stdin;
    } else {
        fprintf(stderr, "- open %s\n", fname);
        fd = fopen(fname, "rb");
        if(!fd) std_err(NULL);
    }
    if(openonly) return(fd);

    if(fd == stdin) {
        xstat.st_size = 0;
        for(filememsz = 0; ; filememsz += len) {
            if(filememsz >= (xstat.st_size + 3)) {  // includes the last NULL too
                xstat.st_size += STDINSZ;
                filemem = realloc(filemem, xstat.st_size);
                if(!filemem) std_err(NULL);
            }
            len = fread(filemem + filememsz, 1, STDINSZ, fd);
            if(len <= 0) break;
        }
    } else {
        fstat(fileno(fd), &xstat);
        filemem = malloc(xstat.st_size + 3);    // add a NULL byte for delimiting the possible text at the end of the file
        if(!filemem) std_err(NULL);
        filememsz = fread(filemem, 1, xstat.st_size, fd);
        fclose(fd);
    }
    filemem[filememsz]     = 0; // at least 3 to NULLify also the unicode strings
    filemem[filememsz + 1] = 0; // it's not needed but I prefer it
    filemem[filememsz + 2] = 0;
    fprintf(stderr, "- read %u bytes from input file\n", filememsz);

    if(parse_PE(0) < 0) {
        if(parse_ELF(0) < 0) std_err("unsupported input file format or architecture");
    }
    return(NULL);
}



FILE *write_file(u8 *fname, int check, int openonly) {
    FILE    *fd;
    u8      ans[16];

    if(!strcmp(fname, "-")) {
        fd = stdout;
    } else {
        fprintf(stderr, "- save %s\n", fname);
        if(check) {
            fd = fopen(fname, "rb");
            if(fd) {
                fclose(fd);
                fprintf(stderr, "- the file %s already exists\n  do you want to overwrite it (y/N)? ", fname);
                fflush(stdin);
                fgets(ans, sizeof(ans), stdin);
                if(tolower(ans[0]) != 'y') exit(1);
            }
        }
        fd = fopen(fname, "wb");
        if(!fd) std_err(NULL);
    }
    if(openonly) return(fd);

    fwrite(filemem, 1, filememsz, fd);
    if(fd != stdout) fclose(fd);
    return(NULL);
}



u8 *strings_filter(int offset, int *type) {
    static int  ubuffsz = 0;
    static u8   *ubuff  = NULL;
    int     i;
    u8      *string,
            *p,
            c;

    if(file2rva(offset) == -1) return(NULL);
    if(type) *type = IS_ASCII;
    string = filemem + offset;

    if(try_unicode && (strlen(string) == 1)) {
        p = string;
        for(i = 0; ; i++) {
            if(!p[0]) break;
            if(p[1]) break;
            if(i >= ubuffsz) {
                ubuffsz += BUFFSZ;
                ubuff = realloc(ubuff, ubuffsz + 1);
                if(!ubuff) std_err(NULL);
            }
            ubuff[i] = p[0];
            p += 2;
        }
        if(p[0] || p[1]) return(NULL);      // the final double NULL delimiter is missed
        ubuff[i] = 0;
        string = ubuff;
        if(type) *type = IS_UNICODE;
    }

    i = 0;
    for(p = string; *p; p++) {
        if((*p != '\r') && (*p != '\n')) i++;
    }
    if(i < minstringlen) return(NULL);

    if(no_filter) return(string);

    if(string[0] == 0xff) return(NULL);

    if(asian_only) {
        if(string[0] < 0x80) return(NULL);  // japanese and other similar strings only
        if(type) *type = IS_ASIAN;
        return(string);
    }

    for(i = 0; (c = string[i]); i++) {
        if(!((c == '\t') || (c == '\r') || (c == '\n') || ((c >= ' ') && (c < 0x7f)))) return(NULL);
    }
    return(string);
}



u8 *string2c(u8 *string, int type) {
    static int  retsz   = 0;
    static u8   *ret    = NULL;
    int     i,
            r;
    u8      c,
            x;

    r = 0;
    for(i = 0; (c = string[i]); i++) {
        if(r >= retsz) {
            retsz += BUFFSZ;
            ret = realloc(ret, retsz + 10); // better to allocate more space
            if(!ret) std_err(NULL);
        }

        if(!i) {
            ret[r++] = '\"';
        }
        switch(c) {
            case '\\': { ret[r++] = '\\'; ret[r++] = '\\';  break; }
            case '\"': { ret[r++] = '\\'; ret[r++] = '\"';  break; }
            case '\'': { ret[r++] = '\\'; ret[r++] = '\'';  break; }
            case '\?': { ret[r++] = '\\'; ret[r++] = '\?';  break; }
            case '\r': { ret[r++] = '\\'; ret[r++] = 'r';   break; }
            case '\n': { ret[r++] = '\\'; ret[r++] = 'n';   break; }
            case '\a': { ret[r++] = '\\'; ret[r++] = 'a';   break; }
            case '\b': { ret[r++] = '\\'; ret[r++] = 'b';   break; }
            case '\f': { ret[r++] = '\\'; ret[r++] = 'f';   break; }
            case '\t': { ret[r++] = '\\'; ret[r++] = 't';   break; }
            case '\v': { ret[r++] = '\\'; ret[r++] = 'v';   break; }
            default:   {
                if((type != IS_ASIAN) && !isprint(c)) {
                    r += sprintf(ret + r, "\\x%02x", c);
                    x = tolower(string[i + 1]);
                    if(((x >= '0') && (x <= '9')) || ((x >= 'a') && (x <= 'f'))) {
                        ret[r++] = '\"';
                        ret[r++] = '\"';
                    }
                } else {
                    ret[r++] = c;
                }
                break;
            }
        }
    }
    ret[r++] = '\"';
    ret[r] = 0;
    return(ret);
}



int mycpy(u8 *dst, u8 *src, int type) { // includes NULL delimiter
    int     i,
            len;
    u8      *d;

    len = strlen(src) + 1;
    if(type == IS_UNICODE) {
        if(dst) {
            d = dst;
            for(i = 0; i < len; i++) {
                *d++ = src[i];
                *d++ = 0;
            }
        }
        len *= 2;
    } else {
        if(dst) memcpy(dst, src, len);
    }
    return(len);
}



void strindex_to_txt(FILE *fd) {
    int     i,
            j,
            x;

    for(i = 0; i < strindexn; i++) {
        if(c_string == 0) {
            if(!quiet) {
                fprintf(fd, "%s%d", MYLINES, strindex[i].string_type);
                for(j = 0; j < strindex[i].offsets; j++) {
                    fprintf(fd, "-%08x", strindex[i].pointer_offset[j]);
                }
                fprintf(fd, "___");
                for(j = 0; j < strindex[i].offsets; j++) {
                    for(x = j + 1; x < strindex[i].offsets; x++) {
                        if(strindex[i].string_offset[x] == strindex[i].string_offset[j]) break;
                    }
                    if(x >= strindex[i].offsets) fprintf(fd, "%08x/%08x.", strindex[i].string_offset[j], strindex[i].string_rva[j]);
                }
                fseek(fd, -1, SEEK_CUR);
                fprintf(fd, FILE_CRLF);
            }
            fprintf(fd, "%s"FILE_CRLF, strindex[i].string);
        } else if(c_string == 1) {
            fprintf(fd, "    %s%s"FILE_CRLF, (strindex[i].string_type == IS_UNICODE) ? "L" : "", string2c(strindex[i].string, strindex[i].string_type));
        } else if(c_string == 2) {
            fprintf(fd, "    { \"\", %s%s },"FILE_CRLF, (strindex[i].string_type == IS_UNICODE) ? "L" : "", string2c(strindex[i].string, strindex[i].string_type));
        }
    }
}



void debug_strindex(void) {
    #ifdef DEBUG
    FILE    *fddbg;
    fddbg = fopen("exestringz_dbg.txt", "wb");
    if(!fddbg) std_err(NULL);
    strindex_to_txt(fddbg);
    fclose(fddbg);
    #endif
}



void strindex_to_bin(FILE *fd) {
    int     i,
            old_len,
            new_len;
    u8      ans[16],
            *old_string;

    for(i = 0; i < strindexn; i++) {    // exestingz doesn't read the string index, so I use pointer_offset[0] instead of string_offset[0]
        if((align_string >= 2) && (strindex[i].pointer_offset[0] % align_string)) continue;

        old_string = strings_filter(strindex[i].pointer_offset[0], NULL);
        if(old_string) {
            old_len = strlen(old_string);
            new_len = strlen(strindex[i].string);
            if(new_len > old_len) {
                fprintf(stderr,
                    "- the new string at offset %08x (%d) is longer than the older one (%d)\n"
                    "  do you want to overwrite this specific string (y/N)? ",
                    strindex[i].pointer_offset[0], new_len, old_len);
                fflush(stdin);
                fgets(ans, sizeof(ans), stdin);
                if(tolower(ans[0]) != 'y') continue;
            }
        }
        mycpy(filemem + strindex[i].pointer_offset[0], strindex[i].string, strindex[i].string_type);
    }

    debug_strindex();
}



void strindex_to_exe(FILE *fd) {
    int     i,
            j,
            string_len,
            new_section_p;

    new_section_p = 0;

    for(i = 0; i < strindexn; i++) {
        string_len = mycpy(NULL, strindex[i].string, strindex[i].string_type);
        while((new_section_p + string_len + 10) >= new_section_sz) {    // NULL delimiters plus padding
            new_section_sz += sections_align;   // repeat in case sections_align is smaller than len
            new_section = realloc(new_section, new_section_sz);
            if(!new_section) std_err(NULL);
        }
        string_len = mycpy(new_section + new_section_p, strindex[i].string, strindex[i].string_type);

        for(j = 0; j < strindex[i].offsets; j++) {
            if((align_string >= 2) && (strindex[i].string_offset[j] % align_string)) continue;
            if(strindex[i].pointer_offset[j] >= (filememsz - sizeof(u32))) continue;
            *(u32 *)(filemem + strindex[i].pointer_offset[j]) = imagebase + new_section_rva + new_section_p;
        }

        for(new_section_p += string_len; new_section_p % 4; new_section_p++) {
            new_section[new_section_p] = 0;                 // 32 bit padding
        }
    }

    debug_strindex();
}



int add_to_strindex(int type, int pos, int str, int rva, u8 *string, int *string_len) {
    int     i,
            j;

    if(string_len) *string_len = 0;
    if((align_string >= 2) && (str % align_string)) return(-1);

    if(!binary_mode) {
        for(i = 0; i < strindexn; i++) {
            if((strindex[i].string_type == type) && !strcmp(strindex[i].string, string)) {  // we have found the same string
                for(j = 0; j < strindex[i].offsets; j++) {
                    if(strindex[i].pointer_offset[j] == pos) return(-1);    // will never happen because this is the eip where is linked the string
                }
                strindex[i].pointer_offset = realloc(strindex[i].pointer_offset, sizeof(u32) * (strindex[i].offsets + 1));
                strindex[i].string_offset  = realloc(strindex[i].string_offset,  sizeof(u32) * (strindex[i].offsets + 1));
                strindex[i].string_rva     = realloc(strindex[i].string_rva,     sizeof(u32) * (strindex[i].offsets + 1));
                strindex[i].pointer_offset[strindex[i].offsets] = pos;
                strindex[i].string_offset[strindex[i].offsets]  = str;
                strindex[i].string_rva[strindex[i].offsets]     = rva;
                strindex[i].offsets++;
                if(string_len) *string_len = mycpy(NULL, string, type);
                return(0);
            }
        }
    }

    if(strindexn >= strindexsz) {
        strindexsz += 2000;
        strindex = realloc(strindex, sizeof(strindex_t) * strindexsz);
        if(!strindex) std_err(NULL);
    }
    strindex[strindexn].offsets           = 1;
    strindex[strindexn].pointer_offset    = malloc(sizeof(u32) * 1);
    strindex[strindexn].string_offset     = malloc(sizeof(u32) * 1);
    strindex[strindexn].string_rva        = malloc(sizeof(u32) * 1);
    strindex[strindexn].pointer_offset[0] = pos;
    strindex[strindexn].string_offset[0]  = str;
    strindex[strindexn].string_rva[0]     = rva;
    strindex[strindexn].string_type       = type;
    strindex[strindexn].string            = strdup(string);
    strindexn++;
    if(string_len) *string_len = mycpy(NULL, string, type);
    return(0);
}



int string_1(FILE *fd) {
    t_disasm da;
    u32     entrypoint_offset = 0;
    int     i,
            eip,
            asm_size,
            pattern,
            limit,
            array_rva,
            array,
            str,
            pos,
            tot,
            type,
            *p,
            section_exe,
            string_len;
    u8      *string;

    if(c_string == 0) {
        if(!quiet) fprintf(fd, "%s\"type\"%c\"offsets where is pointed the string\"_\"offsets/rva of the string\""FILE_CRLF, MYLINES, MYLINESDIV);
    } else if(c_string == 1) {
        // nothing to init
    } else if(c_string == 2) {
        fprintf(fd, "    static const char   *str1_str2[][2] = {"FILE_CRLF);
    }

    tot = 0;
    if(binary_mode) {
        limit = filememsz;
        i = 0;
        while(i <= limit) {
            string = strings_filter(i, &type);
            if(string && !add_to_strindex(type, i, i, -1, string, &string_len)) {
                tot++;
                i += string_len;
            } else {
                i++;
            }
        }
        goto string_1_finish;
    }

    if(scan_entryp) {
        entrypoint_offset = rva2file(entrypoint);
    }

    for(section_exe = 0; section_exe < sections; section_exe++) {
        if(!(section[section_exe].type & (IMAGE_SCN_MEM_EXECUTE | IMAGE_SCN_CNT_CODE))) continue;

        limit = section[section_exe].offset + section[section_exe].size;
        if(limit > filememsz) continue;

        fprintf(stderr, "- scan section \"%s\" from offset %08x rva %08x for %d bytes\n", section[section_exe].name, section[section_exe].offset, file2rva(section[section_exe].offset), section[section_exe].size);

        pattern = section[section_exe].offset;
        if(scan_entryp && ((entrypoint_offset >= pattern) && (entrypoint_offset < limit))) {
            pattern = entrypoint_offset;
            fprintf(stderr, "- entry point scanning activated %08x\n", entrypoint);
        }
        //eip     = imagebase + section[section_exe].rva;
        for(eip = file2rva(pattern); pattern < limit; pattern += asm_size, eip += asm_size) {
            asm_size = Disasm(filemem + pattern, limit - pattern, eip, &da, DISASM_DATA);
            if(asm_size <= 0) break;

            if(!(((da.cmdtype & C_TYPEMASK) == C_CMD) || ((da.cmdtype & C_TYPEMASK) == C_PSH) || ((da.cmdtype & C_TYPEMASK) == C_DAT))) {
                continue;   // fast solution
            }

            for(i = 0; i < 2; i++) {
                if(!i) {
                    array_rva = da.adrconst;    // like mov eax, dword ptr [4*ecx+OFFSET]
                } else {
                    array_rva = da.immconst;    // like mov eax, OFFSET
                }
                if(array_rva <= 0) continue;
                array = rva2file(array_rva);
                if(array <= 0) continue;
                if(array >= filememsz) continue;
                break;
            }
            if(i >= 2) continue;

            // get also the stringz located in the .text field
            //if((array >= section[section_exe].offset) && (array < limit)) {
                //continue;   // a string can't be located in the .text section
            //}

            if(i == 1) {                        // immconst
                pos = pattern + asm_size - 4;   // the last 4 bytes are ever the offset
                str = array;

                string = strings_filter(str, &type);
                if(string) {                    // yeah, don't give up if doesn't contain a string, it could be an array of strings!
                    if(!add_to_strindex(type, pos, str, file2rva(str), string, NULL)) tot++;
                    continue;
                }
            }

            for(p = (int *)(filemem + array); *p && (((u8 *)p - filemem) < filememsz); p++) {
                str = rva2file(*p);
                if(str <= 0) continue;          // don't break, I want ALL them
                if(str >= filememsz) continue;  // don't break, I want ALL them

                string = strings_filter(str, &type);
                if(!string) continue;

                if(c_string == 0) {
                    pos = (u8 *)p - filemem;
                } else {
                    pos = str;
                }
                if(!add_to_strindex(type, pos, str, file2rva(str), string, NULL)) tot++;
            }
        }
    }

string_1_finish:
    strindex_to_txt(fd);

    if(c_string == 0) {
        if(!quiet) fprintf(fd, "%s0"FILE_CRLF, MYLINES);
    } else if(c_string == 1) {
    } else if(c_string == 2) {
        fprintf(fd, "    { NULL, NULL }"FILE_CRLF);
        fprintf(fd, "    };"FILE_CRLF);
    }
    fprintf(stderr, "- %d strings collected\n", tot);
    return(tot);
}



u8 *myfgets(FILE *fd, int *eof, int *crlf) {   // supports both \n and \r\n
    static int  buffsz  = 0;
    static u8   *buff   = NULL;
    int     i,
            c;

    *eof  = 0;
    *crlf = LINE_LF;
    for(i = 0;;) {
        c = fgetc(fd);
        if(c < 0) {
            *eof = 1;
            break;
        }

        if(i >= buffsz) {
            buffsz += BUFFSZ;
            buff = realloc(buff, buffsz + 3);   // includes the NULL byte plus the possible
            if(!buff) std_err(NULL);            // additional crlf added by the caller function
        }

        if(c == '\r') {         // \r\n
            *crlf = LINE_CRLF;
            continue;
        }
        if(c == '\n') break;    // \r\n and \n
        buff[i] = c;
        i++;
    }
    buff[i] = 0;
    return(buff);
}



int string_2(FILE *fd) {
    int     i,
            *pos,
            posn,
            pos_size,
            tot,
            len,
            type,
            string_len,
            string_size,
            delthelast,
            eof,
            crlf;
    u8      *buff,
            *string,
            *p;

    tot         = 0;

    pos         = NULL;
    pos_size    = 0;
    string      = NULL;
    string_size = 0;

    posn        = 0;
    string_len  = 0;
    delthelast  = 0;
    type        = IS_ASCII;

    for(;;) {
        buff = myfgets(fd, &eof, &crlf);    // returns the line without line feed

        p = buff + strlen(buff);
        if(force_crlf) crlf = LINE_CRLF;
        if(crlf == LINE_LF) {
            *p++ = '\n';
        } else if(crlf == LINE_CRLF) {
            *p++ = '\r';
            *p++ = '\n';
        }
        *p = 0;

        if(!memcmp(buff, MYLINES, sizeof(MYLINES) - 1)) {
            last_line_check:    // unfortunately this goto is required for the end of file
            if(posn && (string_len > 0)) {
                string_len -= delthelast;
                string[string_len] = 0;
                for(i = 0; i < posn; i++) {
                    if(!add_to_strindex(type, pos[i], 0, 0, string, NULL)) tot++;
                }
            }

            if(eof) break; // in case someone deleted the last line of strings.txt

            posn        = 0;
            string_len  = 0;
            delthelast  = 0;
            type        = IS_ASCII;
            p = buff + (sizeof(MYLINES) - 1);
            if((sscanf(p, "%d", &type) == 1) && (type != IS_WRONG)) {
                for(posn = 0; (p = strchr(p + 1, MYLINESDIV)); posn++) {
                    if(posn >= pos_size) {
                        pos_size += 32;
                        pos = realloc(pos, sizeof(int) * pos_size);
                        if(!pos) std_err(NULL);
                    }
                    p++;    // skip MYLINESDIV
                    if(sscanf(p, "%08x", &pos[posn]) != 1) break;
                }
            }
        } else {
            len = strlen(buff);
            while((string_len + len + 4) >= string_size) {  // while() is not useful, if() was enough but I want to be sure of that
                string_size += BUFFSZ;
                string = realloc(string, string_size + 1);  // includes the NULL byte
                if(!string) std_err(NULL);
            }
            memcpy(string + string_len, buff, len);
            string_len += len;

            if(crlf == LINE_LF) {
                delthelast = 1;
            } else if(crlf == LINE_CRLF) {
                delthelast = 2;
            }

            if(eof) goto last_line_check;
        }
    }

    if(binary_mode) {
        strindex_to_bin(fd);
    } else {
        strindex_to_exe(fd);
    }

    fprintf(stderr, "- %d strings collected\n", tot);
    return(tot);
}



int main(int argc, char *argv[]) {
    FILE    *fd;
    int     i,
            op,
            tot;
    u8      *fexe,
            *ftxt;

    fputs("\n"
        "Executable's strings lister and replacer "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "  disassembler engine from Oleh Yuschuk\n"
        "\n", stderr);

    if(argc < 4) {
        fprintf(stderr, "\n"
            "Usage: %s [options] <cmd_1/2> <file.EXE/DLL> <strings.TXT>\n"
            "\n"
            "Options for cmd 1 (executable --> strings.txt)\n"
            "-f      disable the chars filtering on the strings, ANY string will be valid\n"
            "        by default the tool collects only the strings composed by chars\n"
            "        between 0-9, a-z, A-Z and the various symbols and punctuation\n"
            "-a      asian strings only, all those which start with a byte higher than 0x7f\n"
            "-u      disable the collecting of unicode strings\n"
            "-l LEN  minimum length of a string for being considered valid (%d)\n"
            "-b      activate the binary mode, the input file will not be disassembled so\n"
            "        the tool will try to retrieve all the existent sequences of bytes\n"
            "        which look like a string, just like the \"strings\" unix program\n"
            "-e      starts the disassembly scanning from the entry point of the file\n"
            "\n"
            "-A NUM  gets only the strings aligned each NUM bytes, for example if NUM is 4\n"
            "        (32 bits) it will collect only the strings which start at offsets like\n"
            "        4, 8, 0x200 or 0x10004 reducing false positives. works with cmd 2 too\n"
            "\n"
            "the following cmd 1 options are ONLY for visualization, because the generated\n"
            "strings.txt will be not compatible with cmd 2\n"
            "-c      C like output with \"line1\\nline2\" \"line3\" and so on\n"
            "-C      as above but located in a C array like: { \"\", \"line\" },\n"
            "-q      quiet, will not display the __________ lines\n"
            "\n"
            /* "Options for cmd 2 (executable <-- strings.txt)\n"
            "-r      force the putting of carriage-return/line-feed end of lines (0d 0a)\n"
            "        in the destination exe instead of the common line-feed only (0a)\n"
            "\n" */
            "use \"-\" as strings.txt for dumping all the strings on the screen\n"
            "the tool supports both PE and ELF executables with cmd 1, only PE with cmd 2\n"
            "steps for creating strings.txt and reputting them back into the executable:\n"
            "\n"
            "  1) generates strings.txt:\n"
            "     %s 1 program.exe strings.txt\n"
            "\n"
            "     edit strings.txt as you want without modifying the __________ lines\n"
            "     you can also remove all the strings you don't need to change, remember\n"
            "     only to remove also the __________ line above each one of them\n"
            "\n"
            "  2) put all the new strings of strings.txt in the executable\n"
            "     %s 2 program.exe strings.txt\n"
            "\n", argv[0], minstringlen, argv[0], argv[0]);
        exit(1);
    }

    argc -= 3;
    for(i = 1; i < argc; i++) {
        if(((argv[i][0] != '-') && (argv[i][0] != '/')) || (strlen(argv[i]) != 2)) {
            fprintf(stderr, "\nError: wrong argument (%s)\n", argv[i]);
            exit(1);
        }
        switch(argv[i][1]) {
            case 'a': asian_only    = 1;                break;
            case 'u': try_unicode   = 0;                break;
            case 'f': no_filter     = 1;                break;
            case 'l': minstringlen  = atoi(argv[++i]);  break;
            case 'A': align_string  = atoi(argv[++i]);  break;
            case 'b': binary_mode   = 1;                break;
            case 'e': scan_entryp   = 1;                break;
            case 'c': c_string      = 1;                break;
            case 'C': c_string      = 2;                break;
            case 'q': quiet         = 1;                break;
            // case 'r': force_crlf          = 1;                break; // useless, nobody needs it
            default: {
                fprintf(stderr, "\nError: wrong argument (%s)\n", argv[i]);
                exit(1);
            }
        }
    }

    for(i = 0; i < 3; i++) {
        if((argv[argc + i][0] == '-') && argv[argc + i][1]) {
            fprintf(stderr, "\n"
                "Error: seems that you have missed one of the needed parameters.\n"
                "       launch this tool without arguments for a quick help\n");
            exit(1);
        }
    }
    op   = atoi(argv[argc]);
    fexe = argv[argc + 1];
    ftxt = argv[argc + 2];

    load_file(fexe, 0);

    if(op == 1) {
        fprintf(stderr, "- executable --> strings\n");
        fd = write_file(ftxt, 1, 1);
        string_1(fd);
        fclose(fd);

    } else if(op == 2) {
        fprintf(stderr, "- strings --> executable\n");
        fd = load_file(ftxt, 1);
        tot = string_2(fd);
        fclose(fd);
        if(tot) {
            parse_PE(1);
            write_file(fexe, 1, 0);
        }

    } else {
        std_err("wrong operation, launch the tool without arguments for the help");
    }

    free_all();
    fprintf(stderr, "- done\n");
    return(0);
}


