/*
    Copyright 2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

#ifdef WIN32
    #include <winsock.h>
    #include <direct.h>
    #include "winerr.h"

    #define close       closesocket
    #define sleep       Sleep
    #define in_addr_t   uint32_t
    #define ONESEC      1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netdb.h>
    #include <sys/ioctl.h>
    #include <net/if.h>

    #define ONESEC      1
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.1.3"
#define BUFFSZ      4000
#define DATASZ      16000
#define MAXWAIT     30
#define DEFAULTUSER "anonymous"
#define DEFAULTPASS "guest@example.com"
#define OUT_        fprintf(stdout,
#define ERR_        fprintf(stderr,



int     verbose     = 0,
        html        = 0;
u8      *fullpath,
        *fullhost,
        *fullurl    = "",
        *vismask    = "S";
typedef struct {
    int     sd;
    int     sdd;
    u8      *user;
    u8      *pass;
    u8      *host;
    u16     port;
    u8      *uri;
    in_addr_t   ip;
    in_addr_t   ipd;
    u16     portd;
    u8      buff[BUFFSZ + 1];
    u8      passive;
    struct sockaddr_in  peerl;
    int     sdl;
    in_addr_t   myip;
} ftp_t;

typedef struct {
    u8      *perm;
    u8      *num;
    u8      *user;
    u8      *group;
    u8      *fsize;
    u8      *month;
    u8      *day;
    u8      *timex;
    u8      *fname;
} ftpinfo_t;



void ftp_host(u8 *url, ftp_t *ftp);
int conna(in_addr_t ip, u16 port);
in_addr_t get_sock_ip_port(int sd, u16 *port);
int binda(int sdl, struct sockaddr_in *peerl);
int ftp_conna(ftp_t *ftp);
u8 *ftp_pasv(ftp_t *ftp, int *size);
int ftp_cmd(ftp_t *ftp, u8 *cmd, u8 *par);
void ftp_quit(ftp_t *ftp);
u8 *next(u8 **data);
int ftp_info(u8 *line, ftpinfo_t *ftpinfo);
void scanna(ftp_t *ftp, u8 *folder);
void ftp_error(ftp_t *ftp);
int timeout(int sd, int secs);
in_addr_t resolv(u8 *host);
void std_err(void);



int main(int argc, char *argv[]) {
    ftp_t   ftp;
    int     i,
            passive = 1;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    // setbuf(stdout, NULL);
    // setbuf(stderr, NULL);

    fputs("\n"
        "myftpidx "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 2) {
        ERR_"\n"
            "Usage: %s [options] <ftp://[user:pass@]host[:port][/path]>\n"
            "\n"
            "Options\n"
            "-v       verbose\n"
            "-f       visualization of the full ftp url instead of only the filename\n"
            "-m MASK  choose the parameters to visualize: Permissions, Files, User, Group,\n"
            "         Size, Month, Day and Time. so -m STP visualizes the file size\n"
            "         followed by its 24h time and its permissions (default is size)\n"
            "-x       HTML output\n"
            "-y       HTML output in cells with border\n"
            "-p       disable passive mode (automatic if needed)\n"
            "\n"
            "Examples:\n"
            "  myftpidx -m \"\" ftp://localhost > filelist.txt\n"
            "  myftpidx -m PFUGSMDT -y ftp://localhost/path/folder > dump.html\n"
            "  myftpidx \"ftp://myuser:mypass@1.2.3.4:1234/pub/folder1/folder2\" > dump.txt\n"
            "\n", argv[0]);
        exit(1);
    }

    argc--;
    for(i = 1; i < argc; i++) {
        if(((argv[i][0] != '-') && (argv[i][0] != '/')) || (strlen(argv[i]) != 2)) {
            ERR_"\nError: wrong argument (%s)\n", argv[i]);
            exit(1);
        }
        switch(argv[i][1]) {
            case '-':
            case '?':
            case 'h': {
                exit(1);
                } break;
            case 'v': {
                verbose     = 1;
                } break;
            case 'f': {
                fullurl     = NULL;
                } break;
            case 'm': {
                vismask     = argv[++i];
                } break;
            case 'x': {
                html        = 1;
                } break;
            case 'y': {
                html        = 2;
                } break;
            case 'p': {
                passive     = 0;
                } break;
            default: {
                ERR_"\nError: wrong argument (%s)\n", argv[i]);
                exit(1);
            }
        }
    }

    ftp_host(argv[argc], &ftp);
    ftp.passive = passive;

    fullpath = malloc(BUFFSZ);
    if(!fullpath) std_err();
    fullpath[0] = 0;

    fullhost = malloc(6 + strlen(ftp.user) + 1 + strlen(ftp.pass) + 1 + strlen(ftp.host) + 1 + 5 + 1);
    if(!fullhost) std_err();
    strcpy(fullhost, "ftp://");
    if(strcmp(ftp.user, DEFAULTUSER) || strcmp(ftp.pass, DEFAULTPASS)) sprintf(fullhost + strlen(fullhost), "%s:%s@", ftp.user, ftp.pass);
    sprintf(fullhost + strlen(fullhost), "%s", ftp.host);
    if(ftp.port != 21) sprintf(fullhost + strlen(fullhost), ":%hu", ftp.port);
    if(!fullurl) fullurl = fullhost;

    vismask = strdup(vismask);
    for(i = 0; vismask[i]; i++) {
        vismask[i] = toupper(vismask[i]);
    }

    if(ftp_conna(&ftp) < 0) std_err();

    if(html) OUT_"<html>\n<style type=\"text/css\"><!--*{font-family:arial;font-size:11px;text-decoration:none}--></style>\n<title>%s</title>\n<b>", fullhost);
    OUT_"indexing of %s\n", fullhost);
    if(html) OUT_"</b><br>\n<table border=%d cellspacing=0 cellpadding=1>\n", html - 1);
    ERR_"- start indexing\n");
    scanna(&ftp, ftp.uri);

    ERR_"- finished\n");
    if(html) OUT_"</table></html>");
    ftp_quit(&ftp);
    free(fullpath);
    return(0);
}



void ftp_host(u8 *url, ftp_t *ftp) {
    u16     port  = 21;
    u8      *host = NULL,
            *uri  = NULL,
            *user = NULL,
            *pass = NULL,
            *p;

    host = strdup(url);

    p = strstr(host, "://");
    if(!p) p = strstr(host, ":\\\\");
    if(p) {
        for(p += 3; *p; p++) {
            if((*p != '/') && (*p != '\\')) break;
        }
        host = p;
    }

    for(p = host; *p; p++) {
        if((*p == '/') || (*p == '\\')) {
            uri = p;
            break;
        }
    }
    if(uri) *uri++ = 0;
    if(!uri) uri = "";

    p = strrchr(host, '@');
    if(p) {
        *p = 0;

        user = host;

        pass = strchr(host, ':');
        if(pass) {
            *pass++ = 0;
        } else {
            pass = "";
        }

        host = p + 1;
    }

    p = strchr(host, ':');
    if(p) {
        *p = 0;
        port = atoi(p + 1);
    }

    if(!user) {
        user = DEFAULTUSER;
        if(!pass) pass = DEFAULTPASS;
    }

    p = uri + strlen(uri) - 1;
    if(*p == '/') *p = 0;

    ftp->sd      = -1;
    ftp->sdd     = -1;
    ftp->host    = host;
    ftp->port    = port;
    ftp->uri     = uri;
    ftp->user    = user;
    ftp->pass    = pass;
    ftp->ip      = INADDR_NONE;
    ftp->ipd     = INADDR_NONE;
    ftp->portd   = 0;
    ftp->buff[0] = 0;
    ftp->passive = 1;
    memset(&ftp->peerl, 0, sizeof(struct sockaddr_in));
    ftp->sdl     = -1;
    ftp->myip    = INADDR_NONE;
}



int conna(in_addr_t ip, u16 port) {
    struct  sockaddr_in peer;
    int     sd,
            i;

    peer.sin_addr.s_addr = ip;
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    if(verbose) ERR_"- connect to %s:%hu...", inet_ntoa(*(struct in_addr *)&ip), port);
    sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sd < 0) std_err();
    for(i = 1; ; i++) {
        if(!connect(sd, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))) break;
        if(i == 3) return(-1);
        ERR_"- new connection %d closed or timed out, retrying within some seconds\n", i);
        sleep(ONESEC);
    }
    if(verbose) ERR_" ok\n");
    return(sd);
}



in_addr_t get_sock_ip_port(int sd, u16 *port) {
    struct sockaddr_in  peer;
    int     psz;

    psz = sizeof(struct sockaddr_in);
    if(getsockname(sd, (struct sockaddr *)&peer, &psz)
      < 0) std_err();
    if(port) *port = peer.sin_port;
    return(peer.sin_addr.s_addr);
}



int binda(int sdl, struct sockaddr_in *peerl) {
    int     sd,
            psz;

    if(!sdl) {
        peerl->sin_addr.s_addr = INADDR_ANY;
        peerl->sin_port        = htons(0);
        peerl->sin_family      = AF_INET;

        if(verbose) ERR_"- bind available port...");
        sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
        if(sd < 0) std_err();
        if(bind(sd, (struct sockaddr *)peerl, sizeof(struct sockaddr_in))
          < 0) std_err();
        if(listen(sd, 1)    // only one connection naturally
          < 0) std_err();
        get_sock_ip_port(sd, &peerl->sin_port);
        if(verbose) ERR_" %hu\n", ntohs(peerl->sin_port));
    } else {
        if(timeout(sdl, MAXWAIT) < 0) return(-1);
        psz = sizeof(struct sockaddr_in);
        sd = accept(sdl, (struct sockaddr *)peerl, &psz);
        if(sd < 0) std_err();
        if(verbose) ERR_"- received connection from %s:%hu\n", inet_ntoa(peerl->sin_addr), ntohs(peerl->sin_port));
    }
    return(sd);
}



int ftp_conna(ftp_t *ftp) {
    int     ret;

    if(ftp->sd >= 0) close(ftp->sd);
    if(ftp->ip == INADDR_NONE) ftp->ip = resolv(ftp->host);

    ERR_"- connect to %s:%hu\n", inet_ntoa(*(struct in_addr *)&ftp->ip), ftp->port);
    ftp->sd = conna(ftp->ip, ftp->port);
    if(ftp->sd < 0) return(-1);

    ERR_"- logging in the server\n");
    ret = ftp_cmd(ftp, NULL, NULL);
    if(ret != 230) {
        if((ret < 0) || (ret != 220)) ftp_error(ftp);
        ret = ftp_cmd(ftp, "USER", ftp->user);
        if((ret != 230) && (ret != 202)) {
            if((ret < 0) || (ret != 331)) ftp_error(ftp);
            ret = ftp_cmd(ftp, "PASS", ftp->pass);
            if((ret < 0) || (ret != 230)) ftp_error(ftp);
        }
    }
    ftp->myip = get_sock_ip_port(ftp->sd, NULL);
    ret = ftp_cmd(ftp, "TYPE", "I");
    if(ret < 0) ftp_error(ftp);
    return(0);
}



u8 *ftp_pasv(ftp_t *ftp, int *size) {
    in_addr_t   myip;
    int     ret,
            len,
            t,
            datasz,
            n1, n2, n3, n4, n5, n6;
    u16     myport;
    u8      portbuff[32],
            *data,
            *p;

    if(!size) {     // PASV
        if(ftp->passive) {
            ret = ftp_cmd(ftp, "PASV", NULL);
            if(ret < 0) ftp_error(ftp);
            if((ret != 227) && (ret != 200)) {
                ftp->passive = 0;
                return(ftp_pasv(ftp, size));
            }

            p = strrchr(ftp->buff, '(');
            if(!p) ftp_error(ftp);
            if(sscanf(p + 1, "%d,%d,%d,%d,%d,%d", &n1, &n2, &n3, &n4, &n5, &n6) != 6) ftp_error(ftp);
            ftp->ipd = htonl((n1 << 24) | (n2 << 16) | (n3 << 8) | (n4));
            if((ftp->ipd == INADDR_ANY) || (ftp->ipd == INADDR_NONE)) ftp->ipd = ftp->ip;
            ftp->portd = (n5 << 8) | (n6);

            if(ftp->sdd >= 0) close(ftp->sdd);
            ftp->sdd = conna(ftp->ipd, ftp->portd);
            if(ftp->sdd < 0) return(ftp_pasv(ftp, size));
        } else {
            ftp->sdl = binda(0, &ftp->peerl);
            if(ftp->sdl < 0) ftp_error(ftp);
            myip   = ntohl(ftp->myip);
            myport = ntohs(ftp->peerl.sin_port);
            sprintf(portbuff, "%d,%d,%d,%d,%d,%d",
                (myip >> 24) & 0xff, (myip >> 16) & 0xff, (myip >> 8) & 0xff, (myip) & 0xff,
                (myport >> 8) & 0xff, (myport) & 0xff);

            ret = ftp_cmd(ftp, "PORT", portbuff);
            if(ret < 0) ftp_error(ftp);
            if((ret != 200)) ftp_error(ftp);
        }
        return(NULL);
    }

    if(!ftp->passive) {
        ftp->sdd = binda(ftp->sdl, &ftp->peerl);
        if(ftp->sdd < 0) ftp_error(ftp);
    }

    data   = NULL;  // data
    datasz = 0;
    for(len = 0; ; len += t) {
        if(len >= datasz) {
            datasz += DATASZ;
            data = realloc(data, datasz);
            if(!data) std_err();
        }
        if(timeout(ftp->sdd, MAXWAIT) < 0) {
            free(data);
            *size = 0;
            return(NULL);
        }
        t = recv(ftp->sdd, data + len, datasz - len, 0);
        if(t <= 0) break;
    }
    close(ftp->sdd);
    ftp->sdd = -1;
    if(!ftp->passive) {
        close(ftp->sdl);
        ftp->sdl = -1;
    }

    data = realloc(data, len + 1);
    if(!data) std_err();
    data[len] = 0;
    *size = len;
    return(data);
}



int ftp_cmd(ftp_t *ftp, u8 *cmd, u8 *par) {
    int     len,
            ret;
    u8      *buff;

    buff = ftp->buff;

    if(cmd) {
        len = snprintf(buff, BUFFSZ, "%s%s%s\r\n", cmd, par ? " " : "", par ? par : (u8 *)"");
        if((len < 0) || (len >= BUFFSZ)) {
            ERR_"- output command too long for the buffer\n");
            return(-1);
        }
        if(verbose) ERR_"> %s", buff);
        if(send(ftp->sd, buff, len, 0) != len) return(-1);
    }

    ret = -1;
    do {
        for(len = 0; len < BUFFSZ; len++) {
            if((timeout(ftp->sd, MAXWAIT) < 0) || (recv(ftp->sd, buff + len, 1, 0) <= 0)) {
                ERR_"- connection lost or timed out, reconnecting automatically\n");
                if(ftp_conna(ftp) < 0) std_err();
                return(ftp_cmd(ftp, cmd, par));
            }
            if(buff[len] == '\r') buff[len] = 0;
            if(buff[len] == '\n') break;
        }
        buff[len] = 0;
        if(verbose) ERR_"< %s\n", buff);
        if(len < 4) continue;
        ret = atoi(buff);
    } while(!ret || (buff[3] == '-'));
    
    memmove(buff, buff + 4, len - 3);
    return(ret);
}



void ftp_quit(ftp_t *ftp) {
    if(ftp->sd < 0) return;
    close(ftp->sd);
    ftp->sd = -1;
}



u8 *next(u8 **data) {
    u8      *ret,
            *p;

    ret = *data;
    p   = *data;
    while(*p && (*p != ' ') && (*p != '\t')) p++;
    if(!*p) return(NULL);
    *p++ = 0;
    while(*p && ((*p == ' ') || (*p == '\t'))) p++;
    *data = p;
    return(ret);
}



int ftp_info(u8 *line, ftpinfo_t *ftpinfo) {
    u8      *p;

    p = line;
    ftpinfo->perm  = next(&p);  if(!ftpinfo->perm)  return(-1);
    ftpinfo->num   = next(&p);  if(!ftpinfo->num)   return(-1);
    ftpinfo->user  = next(&p);  if(!ftpinfo->user)  return(-1);
    ftpinfo->group = next(&p);  if(!ftpinfo->group) return(-1);
    ftpinfo->fsize = next(&p);  if(!ftpinfo->fsize) return(-1);
    ftpinfo->month = next(&p);  if(!ftpinfo->month) return(-1);
    ftpinfo->day   = next(&p);  if(!ftpinfo->day)   return(-1);
    ftpinfo->timex = next(&p);  if(!ftpinfo->timex) return(-1);
    ftpinfo->fname = p;
    return(0);
}



void scanna(ftp_t *ftp, u8 *folder) {
    ftpinfo_t   ftpinfo;
    int     ret,
            i,
            datasz;
    u8      *data,
            *line,
            *currpath,
            *p;

    currpath = fullpath + strlen(fullpath);
    if(folder[0]) {
        if((currpath + strlen(folder) - fullpath) >= BUFFSZ) {
            ERR_"- path too long for the fullpath buffer\n");
            ftp_quit(ftp);
            exit(1);
        }
        sprintf(currpath, "/%s", folder);
    }

    data = ftp_pasv(ftp, NULL);

    ERR_"  %s\n", fullpath);
    ret = ftp_cmd(ftp, "LIST", fullpath);
    if(ret < 0) ftp_error(ftp);
    if((ret != 150) && (ret != 125) && (ret != 200)) goto quit;

    data = ftp_pasv(ftp, &datasz);
    if(!data) {
        ERR_"- data connection lost, retrying automatically\n");
        goto quit_retry;
    }

    ret = ftp_cmd(ftp, NULL, NULL);
    if(ret < 0) ftp_error(ftp);
    if((ret != 226) && (ret != 250) && (ret != 200)) goto quit;

    for(line = data; *line; line = p) {
        for(p = line; *p && (*p != '\n') && (*p != '\r'); p++);
        while(*p && ((*p == '\n') || (*p == '\r'))) *p++ = 0;

        if(ftp_info(line, &ftpinfo) < 0) {
            ERR_"- wrong directories/files informations, retrying automatically\n");
            goto quit_retry;
        }

        if(ftpinfo.perm[0] == 'd') {
            if(!strcmp(ftpinfo.fname, ".") || !strcmp(ftpinfo.fname, "..")) continue;
            scanna(ftp, ftpinfo.fname);

        } else if(ftpinfo.perm[0] == '-') {
            if(html) OUT_"<tr>");

            for(i = 0; vismask[i]; i++) {
                if(html) OUT_"<td>");
                switch(vismask[i]) {
                    case 'P': OUT_"%s ",   ftpinfo.perm);   break;
                    case 'F': OUT_"%4s ",  ftpinfo.num);    break;
                    case 'U': OUT_"%8s ",  ftpinfo.user);   break;
                    case 'G': OUT_"%8s ",  ftpinfo.group);  break;
                    case 'S': OUT_"%12s ", ftpinfo.fsize);  break;
                    case 'M': OUT_"%s ",   ftpinfo.month);  break;
                    case 'D': OUT_"%2s ",  ftpinfo.day);    break;
                    case 'T': OUT_"%5s ",  ftpinfo.timex);  break;
                    default: break;
                }
                if(html) OUT_"</td>");
            }

            if(html) {
                OUT_"<td><a href=\"%s%s/%s\">%s%s/%s</a></td></tr>\n",
                    fullhost, fullpath, ftpinfo.fname,
                    fullurl,  fullpath, ftpinfo.fname);
            } else {
                OUT_"%s%s/%s\n", fullurl, fullpath, ftpinfo.fname);
            }
            fflush(stdout);
        }
    }

quit:
    *currpath = 0;
    if(data) free(data);
    return;
quit_retry:
    *currpath = 0;
    if(data) free(data);
    scanna(ftp, folder);
}



void ftp_error(ftp_t *ftp) {
    ERR_"\n"
        "Error: something wrong during the communication with the server:\n"
        "       %s\n", ftp->buff);
    ftp_quit(ftp);
    exit(1);
}



int timeout(int sd, int secs) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec  = secs;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sd, &fd_read);
    err = select(sd + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



in_addr_t resolv(u8 *host) {
    struct      hostent *hp;
    in_addr_t   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        ERR_"  resolve hostname %s\n", host);
        hp = gethostbyname(host);
        if(!hp) {
            ERR_"\nError: Unable to resolve hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(in_addr_t *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


